import { Injectable, NgModuleFactory, ComponentFactoryResolver, Injector, Compiler, ChangeDetectorRef, NgModule } from '@angular/core';
import { Subject, throwError, Observable, of, forkJoin } from 'rxjs';
import 'reflect-metadata';
import { TotalResultCount, AppGlobal, DataContext, CompanyViews, CompanyWorkdays, CachedViewProperties, CachedViewPropertiesSymbol, CompanyRules, CompanyNotificationMessages, CompanySystemConfiguration, SystemConfiguration, Rule, NotificationMessage, SystemHooks, NotificationVariable, DynamicContentErrorComponent, Views, View, ColumnValues } from 'qmuzik-angular-model-core';
import { HttpClient, HttpParams, HttpHeaders, HttpResponse, HttpErrorResponse } from '@angular/common/http';
import { catchError, map, tap, retry, switchMap } from 'rxjs/operators';
import { Router } from '@angular/router';
import { __awaiter } from 'tslib';

/**
 * All generated data services for standard components inherits from this base service
 * Contains generic/shared functionality as well as overideable interface methods
 */
class BaseService {
    constructor() {
        this.loggedUser = "";
        this.defaults = [];
        this.defaultsSubscription = new Subject();
        this.autoSelectSubscription = new Subject(); //note: indicate to the view that auto-selection check is complete
        this.initialisationService = null;
        this.totalResultCount = new TotalResultCount();
    }
    /**
     * Standard error handler
     *
     * @param error
     */
    handleError(error) {
        let message = "";
        let companyID = error.headers.get("CompanyID");
        let invalidReasonCode = error.headers.get("InvalidReasonCode");
        if (error.error instanceof ErrorEvent) {
            //client-side or network error occurred. Handle it accordingly.
            console.error('An error occurred:', error.error.message);
            message = error.error.message;
        }
        else {
            // The backend returned an unsuccessful response code.
            // The response body may contain clues as to what went wrong,
            console.error(`Backend returned code ${error.status}, ` +
                `body was: ${JSON.stringify(error.error)}`);
            //try different options for the description
            if (error.error.error_description != null) {
                message = error.error.error_description;
            }
            else if (error.error.ModelState != null) {
                for (var key in error.error.ModelState) {
                    if (error.error.ModelState[key])
                        message += error.error.ModelState[key] + '\n';
                }
                ;
            }
            else if (error.error.Message != null) {
                message = error.error.Message + '\r\n' + '\r\n' + error.error.ExceptionMessage;
            }
            else {
                if ((error.error !== undefined) && (error.error != null)) {
                    message = error.error;
                }
                else {
                    message = 'An error has occurred. Please retry.';
                }
            }
            ;
        }
        // return an ErrorObservable with a user-facing error message
        return throwError(message);
    }
    ;
    /**
     * Joins result sets from calls to multiple datacontexts into one dataset
     *
     * @param data
     */
    FormatObserveablesArray(data) {
        let myResults = [];
        for (var i = 0; i < data.length; i++) {
            if ((data[i] !== undefined)
                &&
                    (data[i] !== null))
                for (var j = 0; j < data[i].length; j++) {
                    myResults.push(data[i][j]);
                }
        }
        return myResults;
    }
    /**
     * Perform any formatting required before web API is called upon
     * to save a modelstate
     *
     * @param modelState
     * @param isNew
     */
    FormatBeforeSave(modelState, isNew) {
        let keys = Object.keys(modelState);
        for (var j = 0; j < keys.length; j++) {
            if (modelState[keys[j]] !== null) {
                //note: date formatting
                let isDate = Reflect.getMetadata("IsDate", modelState, keys[j]);
                let isEditable = Reflect.getMetadata("Editable", modelState, keys[j]);
                if (isDate && isEditable) {
                    let datePicker = modelState[keys[j]];
                    let date = new Date(datePicker.year, datePicker.month - 1, datePicker.day);
                    modelState[keys[j]] = date;
                }
            }
            //command values
            if (isNew) {
                let newValueCommand = Reflect.getMetadata("NewValueCommand", modelState, keys[j]);
                if (newValueCommand === "Now") {
                    modelState[keys[j]] = new Date();
                }
            }
            else {
                let updateValueCommand = Reflect.getMetadata("UpdateValueCommand", modelState, keys[j]);
                if (updateValueCommand === "Now") {
                    modelState[keys[j]] = new Date();
                }
                if ((updateValueCommand === "LoggedInUserId") && (this.loggedUser !== "")) {
                    modelState[keys[j]] = this.loggedUser;
                }
            }
        }
        return modelState;
    }
    /**
     *  On a new click we may want to initialise
     *      certain properties to a "variable" value
     *      such as:
     *  TodaysDate
     *  NextWorkDate
     *  TomorrowsDate
     *  LoggedInUser
     *  SystemSettings:<PropertyName>
     *  LoggedInUserDivision:<PropertyName>
     *  LoggedInUserCostCenter:<PropertyName>
     *  EndPoint:<PropertyName> e.g."EndPoint:api/ViewCustomerOrdersHeaderCRUDLibrary/GetMySetting", expects a GET, expects single Parameter "PropertyName"
     *
     * @param modelState
     */
    FormatOnNew(modelState) {
        //note: we want to check if any decorators have been overridden
        //      just a side note for later.
        //note: defaults set up will take precendence over this
        //      it will only apply the setting if it's value is null or undefined 
        try {
            let keys = Object.keys(modelState);
            for (var j = 0; j < keys.length; j++) {
                //command values
                let newValueCommand = Reflect.getMetadata("NewValueCommand", modelState, keys[j]);
                if ((newValueCommand === undefined)
                    ||
                        (newValueCommand === null)) {
                    continue;
                }
                if ((modelState[keys[j]] !== undefined)
                    &&
                        (modelState[keys[j]] !== null)) {
                    continue;
                }
                let commandArray = newValueCommand.split(':');
                if (newValueCommand === "TodaysDate") {
                    modelState[keys[j]] = new Date();
                    continue;
                }
                if (newValueCommand === "TomorrowsDate") {
                    let date = new Date();
                    date.setDate(date.getDate() + 1);
                    modelState[keys[j]] = date;
                    continue;
                }
                if (newValueCommand === "NextWorkDate") {
                    for (var i = 0; i < this.initialisationService.companyWorkdays[modelState.CompanyID].length; i++) {
                        if (this.DaysBetween((new Date(this.initialisationService.companyWorkdays[modelState.CompanyID][i].WorkDate)), new Date()) >= 0) {
                            modelState[keys[j]] = new Date(this.initialisationService.companyWorkdays[modelState.CompanyID][i].WorkDate);
                            break;
                        }
                    }
                    continue;
                }
                if (newValueCommand === "LoggedInUser") {
                    //note: here we must retrieve the logged in user based upon company
                    for (var i = 0; i < this.initialisationService.loginService.dataContext.length; i++) {
                        if (modelState["CompanyID"] === this.initialisationService.loginService.dataContext[i].CompanyID) {
                            modelState[keys[j]] = this.initialisationService.loginService.dataContext[i].EmployeeId;
                            break;
                        }
                    }
                    continue;
                }
                if (commandArray.length === 2) {
                    if (commandArray[0] === "SystemConfiguration") {
                        if (this.initialisationService.companySystemConfiguration[modelState["CompanyID"]]) {
                            modelState[keys[j]] = this.initialisationService.companySystemConfiguration[modelState["CompanyID"]][commandArray[1]];
                        }
                        continue;
                    }
                    if (commandArray[0] === "LoggedInUserDivision") {
                        //note: use initialisation service to retrieve the division information for logged in user
                        continue;
                    }
                    if (commandArray[0] === "LoggedInUserCostCenter") {
                        //note: use initilisation service to retrieve the cost center information for logged in user 
                        continue;
                    }
                }
            }
        }
        catch (e) {
            this.handleError(e);
        }
        return modelState;
    }
    /**
     * Check if any property is a date and editable and convert
     *
     * @param modelState
     */
    FormatBeforeSend(modelState) {
        let keys = Object.keys(modelState);
        for (var j = 0; j < keys.length; j++) {
            if (modelState[keys[j]] !== null) {
                //note: date formatting
                let isDate = Reflect.getMetadata("IsDate", modelState, keys[j]);
                let isEditable = Reflect.getMetadata("Editable", modelState, keys[j]);
                if (isDate && isEditable) {
                    let datePicker = modelState[keys[j]];
                    let date = new Date(datePicker.year, datePicker.month - 1, datePicker.day);
                    modelState[keys[j]] = date;
                }
            }
        }
        return modelState;
    }
    /**
     * Retrieve the operators of an advanced filter
     *
     * @param filter
     * @param advancedFilter
     */
    GetOperatorArray(filter, advancedFilter) {
        let orOperators = [];
        //run through filters,a always add like operator
        //run through each "or" filter
        if (filter !== null) {
            for (var i = 0; i < filter.length; i++) {
                //run trough the properties
                //check if an operator have been set
                //check if a value have been set
                //if so, add the operator
                //else, add null
                //start empty
                let andOperators = [];
                let keys = Object.keys(filter[i]);
                for (var j = 0; j < keys.length; j++) {
                    if (filter[i][keys[j]] === null) {
                        continue;
                    }
                    else {
                        andOperators.push({
                            PropertyName: keys[j],
                            OperatorValue: "*"
                        });
                    }
                }
                orOperators.push(andOperators);
            }
        }
        if ((advancedFilter === undefined) || (advancedFilter === null)) {
            return orOperators;
        }
        //run through each "or" filter
        for (var i = 0; i < advancedFilter.length; i++) {
            //run trough the properties
            //check if an operator have been set
            //check if a value have been set
            //if so, add the operator
            //else, add null
            //start empty
            let andOperators = [];
            let keys = Object.keys(advancedFilter[i]);
            for (var j = 0; j < keys.length; j++) {
                if (advancedFilter[i][keys[j]] === null) {
                    continue;
                }
                let operator = Reflect.getMetadata("Operator", advancedFilter[i], keys[j]);
                if (operator !== undefined) {
                    andOperators.push({
                        PropertyName: keys[j],
                        OperatorValue: operator
                    });
                }
            }
            orOperators.push(andOperators);
        }
        return orOperators;
    }
    /**
     * Retrieve the number of rows that will be requested from the service layer
     *
     * @param key
     * @param modelState
     */
    GetNumberOfRows(key, modelState) {
        let numberOfRows = Reflect.getMetadata("NumberOfRows", modelState, key);
        if ((numberOfRows === undefined) || (numberOfRows === null)) {
            return null;
        }
        else {
            return numberOfRows;
        }
    }
    /**
     * Calculate the difference in days between 2 dates
     *
     * @param date1
     * @param date2
     */
    DaysBetween(date1, date2) {
        //Get 1 day in milliseconds
        let one_day = 1000 * 60 * 60 * 24;
        // Convert both dates to milliseconds
        let date1_ms = date1.getTime();
        let date2_ms = date2.getTime();
        // Calculate the difference in milliseconds
        let difference_ms = date2_ms - date1_ms;
        // Convert back to days and return
        return Math.round(difference_ms / one_day);
    }
    //note: a default implementation so that we don't have to
    //      re-generate all our code when adding method to data service interface
    //#region override child
    DoInstanceActions(modelState, isNew, instance, isList) {
        return new Observable(null);
    }
    GetDefaultsInstance(instanceId, symbolInstanceId, CompanyIDPassed) {
        return new Observable(null);
    }
    GetSymbolInstanceSavedObjectCount(instanceInformation, companyIDPassed) {
        return new Observable(null);
    }
    AddToInstance(modelState) {
        return new Observable(null);
    }
    DeleteInstanceObjects(instance, companyID) {
        return new Observable(null);
    }
    GetPrimaryKeyObject(values, companyID, serializationLevel) {
        return new Observable(null);
    }
    /**
     * Check if an object is of a certain type
     * @param item
     */
    IsType(item) {
        return false;
    }
    GetAutoSelect(instance, modelState, serializationLevel) {
    }
}
BaseService.decorators = [
    { type: Injectable }
];
BaseService.ctorParameters = () => [];

/**
 * General communication between components to manage the application
 */
class AppManagementService extends BaseService {
    constructor(http, globals) {
        super();
        this.http = http;
        this.globals = globals;
        this.leftRegionExpanded = false;
        this.sliderValue = 5; //note: always opens initially with value 5
        this.openDynamicViewRequest = new Subject();
        //note: global error handling e.g. displayed outside context of view
        this.globalErrors = [];
        this.globalWarnings = [];
        this.globalInformations = [];
        this.openMenuSubscription = new Subject();
        this.sliderValueChangedSubject = new Subject();
        this.menuOpen = false;
    }
    /**
     * Retrieve system settings
     */
    getSystemSettings() {
        return this.http.get(this.globals.apiURL + "api/SystemSettings/")
            .pipe(catchError(this.handleError));
    }
    /**
     * Add a new global error (only if it hasn't been logged)
     * @param notificationMessage
     */
    AddGlobalError(notificationMessage) {
        for (var i = 0; i < this.globalErrors.length; i++) {
            if (this.globalErrors[i].ErrorMessage === notificationMessage.ErrorMessage) {
                return;
            }
        }
        this.globalErrors.push(notificationMessage);
    }
    /**
     * Add a new global information message (only if it has not been logged)
     * @param notificationMessage
     */
    AddGlobalInformation(notificationMessage) {
        for (var i = 0; i < this.globalInformations.length; i++) {
            if (this.globalInformations[i].ErrorMessage === notificationMessage.ErrorMessage) {
                return;
            }
        }
        this.globalInformations.push(notificationMessage);
    }
    /**
     * Send a notification the the menu has been openend
     */
    openMenu() {
        this.openMenuSubscription.next();
    }
}
AppManagementService.decorators = [
    { type: Injectable }
];
AppManagementService.ctorParameters = () => [
    { type: HttpClient },
    { type: AppGlobal }
];

class LoginService extends BaseService {
    constructor(http, appGlobal, router, appManagementService) {
        //constructor of inherited class must have this call 
        super();
        this.http = http;
        this.appGlobal = appGlobal;
        this.router = router;
        this.appManagementService = appManagementService;
        this.isLoggedIn = false;
        this.notifyLoginChange = new Subject();
        //note: contains all that is connected e.g. can be used for THIS product
        this.dataContext = [];
        //note: contains all that is not connected, but active for THIS product
        this.dataContextNotConnected = [];
        //note: contains all data contexts for NOT FOR THIS product
        this.dataContextDifferentProduct = [];
        this.defaultCompanyID = "";
        this.dataContextSubscription = new Subject();
        this.dataContextIsActiveCount = new Subject();
        this.dataContextIsSelectedCount = new Subject();
        //the login route target once you've logged in.
        this.loginRouteTarget = "blank";
    }
    /**
     * Return the data contexts where the logged in user has access
     * User will have access if it exists in company views
     *
     * @param internalName
     */
    GetValidDataContexts(internalName) {
        let validDataContext = [];
        for (var i = 0; i < this.dataContext.length; i++) {
            if (this.initialisationService.companyViews[this.dataContext[i].CompanyID]
                &&
                    (this.dataContext[i].CompanyID !== "")) {
                let views = this.initialisationService.companyViews[this.dataContext[i].CompanyID];
                for (var j = 0; j < Object.keys(views).length; j++) {
                    let view = views[Object.keys(views)[j]];
                    if (view.InternalName === internalName) {
                        //note: found, it is valid for the search
                        //      move to the next data context
                        validDataContext.push(this.dataContext[i]);
                        break;
                    }
                }
            }
        }
        return validDataContext;
    }
    validateRegistrationInviteToken(Token, Email) {
        let apiUrl = this.appGlobal.apiURL + 'api/ProductSettings/ValidateInviteToken';
        let httpParams = new HttpParams();
        httpParams = httpParams.append('inviteToken', Token);
        httpParams = httpParams.append('emailAddress', Email);
        return this.http.get(apiUrl, {
            params: httpParams
        })
            .pipe(catchError(this.handleError));
    }
    //note: call the oauth registration endpoint to register new users
    //      users register themselves in the authenticate database
    //      this will allow them to query the data contexts that have been
    //      set up for them
    inviteUsers(inviteProductData) {
        if (this.selectedDataContext.CompanyID === "") {
            return;
        }
        inviteProductData.InvitedBy = this.selectedDataContext.RealEmployeeId;
        let apiUrl = this.appGlobal.apiURL + 'api/ProductSettings/InviteUser';
        return this.http.post(apiUrl, inviteProductData)
            .pipe(catchError(this.handleError));
    }
    SubscribeToNewDataContext(emailAddress, inviteToken) {
        let apiUrl = this.appGlobal.apiURL + 'api/ProductSettings/SubscribeToNewDataContext';
        let httpParams = new HttpParams();
        httpParams = httpParams.append('inviteToken', inviteToken);
        httpParams = httpParams.append('emailAddress', emailAddress);
        return this.http.get(apiUrl, {
            params: httpParams
        })
            .pipe(catchError(this.handleError));
    }
    //note: call the oauth registration endpoint to register new users
    //      users register themselves in the authenticate database
    //      this will allow them to query the data contexts that have been
    //      set up for them
    registerUser(username, email, password, retypePassword, inviteToken) {
        let registrationData = new RegistrationData();
        registrationData.Email = email;
        registrationData.Username = username;
        registrationData.Password = password;
        registrationData.ConfirmPassword = retypePassword;
        registrationData.InviteToken = inviteToken;
        let apiUrl = this.appGlobal.apiURL + 'api/Account/Register';
        return this.http.post(apiUrl, registrationData)
            .pipe(catchError(this.handleError));
    }
    loginExternalProvider(username, provider, authToken) {
        let headers = new HttpHeaders();
        headers = headers.set('Content-Type', 'application/x-www-form-urlencoded');
        let url = this.appGlobal.apiURL + 'Token';
        return this.http.post(url, ("grant_type=externalToken&username=" + username + "&externalToken=" + authToken + "&provider=" + provider), {
            headers: headers
        })
            .pipe(
        //don't need to tap here, don't want to do anything with it
        catchError(this.handleError));
    }
    verifyUser(userId, code) {
        let apiUrl = this.appGlobal.apiURL + 'api/Account/Verify';
        return this.http.post(apiUrl, { UserId: userId, VerificationCode: code })
            .pipe(catchError(this.handleError));
    }
    resetPassword(userId, code, newPassword, retypePassword) {
        let apiUrl = this.appGlobal.apiURL + 'api/Account/ResetPassword';
        return this.http.post(apiUrl, { UserId: userId, ResetToken: code, NewPassword: newPassword, ConfirmPassword: retypePassword })
            .pipe(catchError(this.handleError));
    }
    forgotPassword(username) {
        let apiUrl = this.appGlobal.apiURL + 'api/Account/ForgotPassword';
        return this.http.post(apiUrl, { Username: username })
            .pipe(catchError(this.handleError));
    }
    //note: token retrieved from authentication service
    //allows the user to select datacontexts available
    getAuthenticateBearerToken(username, password) {
        let headers = new HttpHeaders();
        headers = headers.set('Content-Type', 'application/x-www-form-urlencoded');
        let url = this.appGlobal.apiURL + 'Token';
        return this.http.post(url, ("grant_type=password&username=" + username + "&password=" + password), {
            headers: headers
        })
            .pipe(
        //don't need to tap here, don't want to do anything with it
        catchError(this.handleError));
    }
    //note: i just want to succeed in receiving the bearer token from the web api Token endpoint
    //      this bearer token will then be persisted for use during this login session
    getBearerToken(dataContext) {
        //this will come from the authorisation database
        this.defaultCompanyID = "";
        let url = dataContext.ServiceUri + 'api/Accounts';
        let httpParams = new HttpParams();
        httpParams = httpParams.append('ClientName', dataContext.CompanyName);
        httpParams = httpParams.append('Password', dataContext.ServicePassword);
        httpParams = httpParams.append('UserId', dataContext.EmployeeId);
        return this.http.get(url, {
            params: httpParams
        })
            .pipe(catchError(this.handleError));
    }
    //note: on expiry of token this method will retrieve a new token
    //      with the refresh token
    getTokenWithRefreshToken(dataContext) {
        let url = dataContext.ServiceUri + 'api/Account/AuthoriseWithRefreshToken';
        let httpParams = new HttpParams();
        httpParams = httpParams.append('ExpiredToken', dataContext.Token);
        httpParams = httpParams.append('RefreshToken', dataContext.RefreshToken);
        return this.http.get(url, {
            params: httpParams
        })
            .pipe(catchError(this.handleError));
    }
    //logout
    logout() {
        //note: clear the local storage
        localStorage.removeItem('authData');
        //note: if user log's out auto-login is cleared
        localStorage.removeItem('authDataAuthentication');
        this.appManagementService.leftRegionExpanded = false;
        this.initialisationService.cachedViewRules = {};
        this.initialisationService.cachedViewNotifications = {};
        this.initialisationService.cachedViewProperties = {};
        this.initialisationService.cachedViewPropertiesSymbol = {};
        //set logout flag
        this.isLoggedIn = false;
        this.dataContext = [];
        this.notifyLoginChange.next(this.isLoggedIn);
        this.router.navigate(['/login']);
    }
    //note: when the company id is returned from the backend set the selected context
    setDataContextWithCompanyId(companyId) {
        this.defaultCompanyID = companyId;
        for (var i = 0; i < this.dataContext.length; i++) {
            if (companyId === this.dataContext[i].CompanyID) {
                this.selectedDataContext = this.dataContext[i];
                this.dataContextSubscription.next(this.selectedDataContext);
                break;
            }
        }
    }
    // update the DataContext
    updateDataContext(dataContext) {
        let apiUrl = this.appGlobal.apiURL + 'api/DataContextViews/';
        let httpParamss = new HttpParams();
        httpParamss = httpParamss.append('id', dataContext.CompanyID);
        return this.http.put(apiUrl, dataContext, {
            params: httpParamss
        })
            .pipe(catchError(this.handleError));
    }
    //note: a data context is no longer connected
    //      and must be disabled
    DisableDatacontext(dataContext, notificationMessage) {
        //note: first a check to see if there is actually any datacontexts available
        //      if none e.g. only the one we are disabling then we will need to log you out
        //      remember the "" datacontext will always be there, that's why the check starts at 2
        if (this.dataContext.length == 2) {
            alert(notificationMessage.ErrorMessage);
            this.logout();
            return;
        }
        //find the index so we can splice
        for (var i = 0; i < this.dataContext.length; i++) {
            if (this.dataContext[i].DataContextId === dataContext.DataContextId) {
                this.dataContext.splice(i, 1);
                dataContext.ConnectionStatus = "Not Connected";
                dataContext.Token = null;
                dataContext.RefreshToken = null;
                dataContext.ServiceUri = null;
                this.dataContextNotConnected.push(dataContext);
                break;
            }
        }
        //note: notify of change to selected data context
        //note: we have more than 1 connected
        if (this.dataContext.length > 2) {
            this.selectedDataContext = undefined;
            this.defaultCompanyID = "";
            this.dataContextSubscription.next(undefined);
        }
        //note: we have exactly 1 connected
        //      1st one will always be the "" datacontext
        if (this.dataContext.length == 2) {
            this.dataContext[1].IsSelected = true;
            this.selectedDataContext = this.dataContext[1];
            this.defaultCompanyID = this.dataContext[1].CompanyID;
            this.dataContextSubscription.next(this.dataContext[1]);
        }
    }
}
LoginService.decorators = [
    { type: Injectable }
];
LoginService.ctorParameters = () => [
    { type: HttpClient },
    { type: AppGlobal },
    { type: Router },
    { type: AppManagementService }
];
//sent to the oauth registration endpoint
class RegistrationData {
}

/**
 * Services for anyone required to work with data contexts
 */
class DataContextService extends BaseService {
    constructor(http, appGlobal, loginService) {
        //constructor of inherited class must have this call 
        super();
        this.http = http;
        this.appGlobal = appGlobal;
        this.loginService = loginService;
    }
    //note: retrieve all active data contexts from the authentication service
    GetDataContext(username) {
        this.username = username;
        let uri = this.appGlobal.apiURL + 'api/DataContextViews/GetDataContextViewsByUserName';
        let httpParams = new HttpParams();
        httpParams = httpParams.append('userName', username);
        return this.http.get(uri, {
            params: httpParams
        })
            .pipe(catchError(this.handleError));
    }
    //note: retrieve a specific data context from the authentication service
    GetDataContextById(id) {
        let uri = this.appGlobal.apiURL + 'api/DataContextViews/';
        let httpParams = new HttpParams();
        httpParams = httpParams.append('id', id);
        return this.http.get(uri, {
            params: httpParams
        })
            .pipe(catchError(this.handleError));
    }
    //note: return an observable of datacontexts (for all companies <> "")
    GetRetrievedDataContexts() {
        let dataContextFiltered = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID !== "") {
                dataContextFiltered.push(this.loginService.dataContext[i]);
            }
        }
        return of(dataContextFiltered);
    }
    //note : selecting a object in the typeahead we need to
    //        "cast" it to the required type so the instanceof works
    AssignToType(item) {
        //must use assign here, otherwise instancof checks will not work
        return Object.assign(new DataContext(), item);
    }
    //note:  locally run through datacontexts
    //       return the datacontext corresponding to the specified
    //       company id
    GetDataContextForCompanyId(companyID) {
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === companyID) {
                return this.loginService.dataContext[i];
            }
        }
        //not found
        return null;
    }
}
DataContextService.decorators = [
    { type: Injectable }
];
DataContextService.ctorParameters = () => [
    { type: HttpClient },
    { type: AppGlobal },
    { type: LoginService }
];

class InitialisationService extends BaseService {
    constructor(loginService, http, appGlobal, router) {
        super();
        this.loginService = loginService;
        this.http = http;
        this.appGlobal = appGlobal;
        this.router = router;
        this.companyViews = new CompanyViews();
        this.companyWorkdays = new CompanyWorkdays();
        this.cachedViewProperties = new CachedViewProperties();
        this.cachedViewPropertiesSymbol = new CachedViewPropertiesSymbol();
        this.cachedViewRules = new CompanyRules();
        this.cachedViewNotifications = new CompanyNotificationMessages();
        this.companySystemConfiguration = new CompanySystemConfiguration();
        //note: we expose this so the our validation base classes can inject it into Rule scripts
        this.httpClient = http;
        //note: cannot inject in login service...circular reference
        //      login service requires access to companyViews
        this.loginService.initialisationService = this;
    }
    //note: we will use this resolver to retrieve all the initialisation data for a route/component
    resolve(route, state) {
        return __awaiter(this, void 0, void 0, function* () {
            let currentNav = this.router.getCurrentNavigation();
            let viewInternalName;
            for (var i = 0; i < route.data['InternalName'].length; i++) {
                //note: a different version from the one registered in app.module
                //      we may have to mod this is more than one view
                if (currentNav.extras && currentNav.extras.state && currentNav.extras.state.internalName) {
                    viewInternalName = currentNav.extras.state.internalName;
                }
                else {
                    viewInternalName = route.data['InternalName'][i];
                }
                if (this.cachedViewRules[viewInternalName] === undefined) {
                    yield this.GetRules(viewInternalName).toPromise().then((result) => {
                        this.InitialiseCompanyRules(result, viewInternalName);
                    });
                }
                ;
                if (this.cachedViewNotifications[viewInternalName] === undefined) {
                    yield this.GetNotificationMessages(viewInternalName).toPromise().then((result) => {
                        this.InitialiseCompanyNotifications(result, viewInternalName);
                    });
                }
                ;
                if (this.cachedViewProperties[viewInternalName] === undefined) {
                    yield this.GetPropertyConfiguration(viewInternalName).toPromise().then((result) => {
                        this.cachedViewProperties[viewInternalName] = result;
                    });
                    //note: this view is loaded as part of an instance
                    //      get additional properties for the view per symbol template
                    if (currentNav.extras.state.SymbolInstanceId !== undefined && currentNav.extras.state.SymbolInstanceId !== null) {
                        if (currentNav.extras.state.SymbolInstanceId > 0) {
                            yield this.GetPropertyConfigurationSymbol(viewInternalName, currentNav.extras.state.SymbolInstanceId).toPromise().then((result) => {
                                this.cachedViewPropertiesSymbol[(viewInternalName + currentNav.extras.state.SymbolInstanceId.toString())] = result;
                            });
                        }
                    }
                }
            }
            return true;
        });
    }
    //note: retrieve the working days for all active data contexts
    //      we might want to limit result retrieved at some point (this can be done in the backend)
    GetCompanyWorkdays() {
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/Company_CalendarLibrary/GetCompany_CalendarFuture", {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observable.subscribe((result) => {
                if (result.length > 0) {
                    this.companyWorkdays[result[0].CompanyID] = result;
                }
            });
        }
    }
    //note: we must be able to retrieve this for individual companies
    GetCompanyWorkdaysForCompany(companyID) {
        return __awaiter(this, void 0, void 0, function* () {
            let serviceUri = "";
            for (var i = 0; i < this.loginService.dataContext.length; i++) {
                if (this.loginService.dataContext[i].CompanyID === companyID) {
                    serviceUri = this.loginService.dataContext[i].ServiceUri;
                    break;
                }
            }
            let headers = new HttpHeaders({
                "CompanyID": companyID
            });
            let observable = this.http.get(serviceUri + "api/Company_CalendarLibrary/GetCompany_CalendarFuture", {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            yield observable.toPromise().then((result) => {
                if (result.length > 0) {
                    this.companyWorkdays[companyID] = result;
                }
            });
        });
    }
    //note: retrieve the erp system configuration settings per company
    GetSystemConfigurationSettings() {
        let observables = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/SystemConfigurationSettingLibrary", {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            // bit
            // char
            // decimal(12,3)
            // decimal(13,2)
            // decimal(4,2)
            // decimal(5,2)
            // decimal(5,5)
            // decimal(7,2)
            // smalldatetime
            // smallint
            // tinyint
            // varchar(10)
            // varchar(15)
            // varchar(255)
            // varchar(6)
            observable.subscribe((result) => {
                for (var i = 0; i < result.length; i++) {
                    if (this.companySystemConfiguration[result[i].CompanyID] === undefined) {
                        this.companySystemConfiguration[result[i].CompanyID] = new SystemConfiguration();
                    }
                    this.companySystemConfiguration[result[i].CompanyID][result[i].ConfigField.replace(/_/g, '')] = [result[i].ConfigValue][0];
                }
            });
        }
    }
    //note: retrieve the erp system configuration settings for individual companies
    GetSystemConfigurationSettingsForCompany(companyID) {
        return __awaiter(this, void 0, void 0, function* () {
            let serviceUri = "";
            for (var i = 0; i < this.loginService.dataContext.length; i++) {
                if (this.loginService.dataContext[i].CompanyID === companyID) {
                    serviceUri = this.loginService.dataContext[i].ServiceUri;
                    break;
                }
            }
            let headers = new HttpHeaders({
                "CompanyID": companyID
            });
            let observable = this.http.get(serviceUri + "api/SystemConfigurationSettingLibrary", {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            yield observable.toPromise().then((result) => {
                for (var i = 0; i < result.length; i++) {
                    if (this.companySystemConfiguration[companyID] === undefined) {
                        this.companySystemConfiguration[companyID] = new SystemConfiguration();
                    }
                    this.companySystemConfiguration[companyID][result[i].ConfigField.replace(/_/g, '')] = [result[i].ConfigValue][0];
                }
            });
        });
    }
    //note: this will return all views that must be available in a user's menu
    //      it will be returned per company
    //      question: what if you have access on one company to a view, but not another
    GetMyViews() {
        let observables = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/GetMyViews/GetGetMyViews", {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observables.push(observable);
        }
        return forkJoin(observables).pipe(map((data) => {
            return this.FormatObserveablesArray(data);
        }));
    }
    //note: this will return all views that must be available in a user's menu
    //      it will be returned per company
    //      question: what if you have access on one company to a view, but not another
    GetMyViewsForCompany(companyID) {
        let serviceUri = "";
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === companyID) {
                serviceUri = this.loginService.dataContext[i].ServiceUri;
                break;
            }
        }
        let headers = new HttpHeaders({
            "CompanyID": companyID
        });
        let observable = this.http.get(serviceUri + "api/GetMyViews/GetGetMyViews", {
            headers: headers
        })
            .pipe(catchError(this.handleError));
        return observable;
    }
    //note: retrieve the configuration for a view
    GetPropertyConfiguration(viewName) {
        let httpParams = new HttpParams();
        httpParams = httpParams.append('internalName', viewName);
        let observables = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/GetMyViewProperties/GetGetMyViewPropertiesByInternalName", {
                params: httpParams,
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observables.push(observable);
        }
        return forkJoin(observables).pipe(map((data) => {
            return this.FormatObserveablesArray(data);
        }));
    }
    //note: retrieve the configuration for a view for a specific company
    GetPropertyConfigurationForCompany(viewName, companyID) {
        let serviceUri = "";
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === companyID) {
                serviceUri = this.loginService.dataContext[i].ServiceUri;
                break;
            }
        }
        let httpParams = new HttpParams();
        httpParams = httpParams.append('internalName', viewName);
        let headers = new HttpHeaders({
            "CompanyID": companyID
        });
        let observable = this.http.get(serviceUri + "api/GetMyViewProperties/GetGetMyViewPropertiesByInternalName", {
            params: httpParams,
            headers: headers
        })
            .pipe(catchError(this.handleError));
        return observable;
    }
    //note: retrieve the configuration for a view connected to a symbol in qt
    GetPropertyConfigurationSymbol(viewName, symbolInstanceId) {
        let httpParams = new HttpParams();
        httpParams = httpParams.append('internalName', viewName);
        let observables = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID,
                "InstanceInformation": ['0', symbolInstanceId.toString(), '0', '', 'False', '0']
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/GetMyViewProperties/GetGetMyViewPropertiesBySymbolTemplateId", {
                params: httpParams,
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observables.push(observable);
        }
        return forkJoin(observables).pipe(map((data) => {
            return this.FormatObserveablesArray(data);
        }));
    }
    //note: retrieve the rules for a view
    //      this must be optimised. if retrieved once for a session it
    //      won't be retreived again, just returned
    GetRules(viewName) {
        //note: if in production no rules will EVER be allowed to be "eval'ed": security risk + minification issues
        if (this.appGlobal.production) {
            return new Observable(null);
        }
        let httpParams = new HttpParams();
        httpParams = httpParams.append('internalName', viewName);
        let observables = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/GetMyViewRules/GetGetMyViewRulesByInternalName", {
                params: httpParams,
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observables.push(observable);
        }
        return forkJoin(observables).pipe(map((data) => {
            return this.FormatObserveablesArray(data);
        }));
    }
    //note: retrieve the rules for a view for a specific company
    GetRulesForCompany(viewName, companyID) {
        //note: if in production no rules will EVER be allowed to be "eval'ed": security risk + minification issues
        // if(this.appGlobal.production){
        //   return new Observable(null);
        // }
        let serviceUri = "";
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === companyID) {
                serviceUri = this.loginService.dataContext[i].ServiceUri;
                break;
            }
        }
        let httpParams = new HttpParams();
        httpParams = httpParams.append('internalName', viewName);
        let observables = [];
        let headers = new HttpHeaders({
            "CompanyID": companyID
        });
        let observable = this.http.get(serviceUri + "api/GetMyViewRules/GetGetMyViewRulesByInternalName", {
            params: httpParams,
            headers: headers
        })
            .pipe(catchError(this.handleError));
        return observable;
    }
    //note: retrieve the notifications for a view
    GetNotificationMessages(viewName) {
        let httpParams = new HttpParams();
        httpParams = httpParams.append('internalName', viewName);
        let observables = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/GetMyViewNotifications/GetGetMyViewNotificationsByInternalName", {
                params: httpParams,
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observables.push(observable);
        }
        return forkJoin(observables).pipe(map((data) => {
            //#region ensure "no results" will also have something in the cache
            for (var i = 0; i < this.loginService.dataContext.length; i++) {
                if (this.loginService.dataContext[i].CompanyID === "") {
                    continue;
                }
                //note: add the company
                if (this.cachedViewNotifications[this.loginService.dataContext[i].CompanyID] === undefined) {
                    this.cachedViewNotifications[this.loginService.dataContext[i].CompanyID] = {};
                }
                //note: add the view
                if (this.cachedViewNotifications[this.loginService.dataContext[i].CompanyID][viewName] === undefined) {
                    this.cachedViewNotifications[this.loginService.dataContext[i].CompanyID][viewName] = {};
                }
            }
            //#endregion ensure "no results" will also have something in the cache
            return this.FormatObserveablesArray(data);
        }));
    }
    //note: retrieve the notifications for a view
    GetNotificationMessagesForCompany(viewName, companyID) {
        let serviceUri = "";
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === companyID) {
                serviceUri = this.loginService.dataContext[i].ServiceUri;
                break;
            }
        }
        let httpParams = new HttpParams();
        httpParams = httpParams.append('internalName', viewName);
        let headers = new HttpHeaders({
            "CompanyID": companyID
        });
        let observable = this.http.get(serviceUri + "api/GetMyViewNotifications/GetGetMyViewNotificationsByInternalName", {
            params: httpParams,
            headers: headers
        })
            .pipe(catchError(this.handleError));
        return observable;
    }
    //note: retrieve the notification messagas for the system
    GetNotificationMessagesSystem() {
        let observables = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/GetMyViewNotifications/GetGetMyViewNotificationsByIsSystem", {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observables.push(observable);
        }
        return forkJoin(observables).pipe(map((data) => {
            return this.FormatObserveablesArray(data);
        }));
    }
    //note: retrieve the notification messagas for a specified company
    GetNotificationMessagesSystemForCompany(companyID) {
        return __awaiter(this, void 0, void 0, function* () {
            let serviceUri = "";
            for (var i = 0; i < this.loginService.dataContext.length; i++) {
                if (this.loginService.dataContext[i].CompanyID === companyID) {
                    serviceUri = this.loginService.dataContext[i].ServiceUri;
                    break;
                }
            }
            let headers = new HttpHeaders({
                "CompanyID": companyID
            });
            let observable = this.http.get(serviceUri + "api/GetMyViewNotifications/GetGetMyViewNotificationsByIsSystem", {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            yield observable.toPromise().then((result) => {
                this.InitialiseCompanyNotifications(result, 'system');
            });
        });
    }
    //note: initialise the default viewstate properties
    //      on creation of a new instance of the component
    Initialise(viewName, symbolInstanceId) {
        let viewState = {};
        //call on the rest api to retrieve all the control state
        //properties set up in the database for this view
        for (var i = 0; i < this.viewProperties.length; i++) {
            if (viewState[this.viewProperties[i].Company_ID] === undefined) {
                viewState[this.viewProperties[i].Company_ID] = {};
            }
            if (viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName] === undefined) {
                viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName] = {};
            }
            if (viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] === undefined) {
                if (this.viewProperties[i].DataType === "boolean") {
                    if (this.viewProperties[i].Value === "1") {
                        viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = true;
                    }
                    else {
                        viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = false;
                    }
                }
                else if ((this.viewProperties[i].DataType === "number") || (this.viewProperties[i].DataType === "int")) {
                    //note: get rid of trailing 0 decimals by 1st converting to string, then back to float
                    let num;
                    if (this.viewProperties[i].Value) {
                        num = parseFloat(this.viewProperties[i].Value.toString());
                    }
                    viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = num;
                }
                else {
                    viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = this.viewProperties[i].Value;
                }
            }
            //#region symbol properties override
            // if(symbolInstanceId !== undefined && symbolInstanceId !== null && symbolInstanceId > 0){
            //   for(var j = 0; j < this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())].length; j++){
            //     if(
            //       this.viewProperties[i].Company_ID === this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID
            //         &&
            //       this.viewProperties[i].PropertyName === this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName   )
            //     {
            //       //note: we found a property on the symbol that we should override on standard
            //       if(this.viewProperties[i].DataType === "boolean"){
            //         if(this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Value === "1"){
            //           viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = true;
            //         }
            //         else{
            //           viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = false;
            //         }
            //       }
            //       else 
            //         if(this.viewProperties[i].DataType === "number"){
            //           let num: number = this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Value;
            //           viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = num;
            //         }
            //       else
            //         {
            //           viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Value;
            //         }  
            //       break;  
            //     }
            //   }
            // }
            //#endregion symbol properties override
        }
        //#region symbol properties
        if (symbolInstanceId !== undefined && symbolInstanceId !== null && symbolInstanceId > 0) {
            for (var j = 0; j < this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())].length; j++) {
                //note: property name
                if (viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName] === undefined) {
                    viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName] = {};
                }
                //note: property key
                if (viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] === undefined) {
                    viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = {};
                }
                //note: we found a property on the symbol that we should override on standard
                if (this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].DataType === "boolean") {
                    if (this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Value === "1") {
                        viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = true;
                    }
                    else {
                        viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = false;
                    }
                }
                else if (this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].DataType === "number") {
                    let num = this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Value;
                    viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = num;
                }
                else {
                    viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Value;
                }
            }
        }
        //#endregion symbol properties
        return viewState;
    }
    //note: initialise the default viewstate properties by adding on to existing view state
    InitialiseWithViewState(viewName, symbolInstanceId, viewState) {
        //call on the rest api to retrieve all the control state
        //properties set up in the database for this view
        for (var i = 0; i < this.viewProperties.length; i++) {
            if (viewState[this.viewProperties[i].Company_ID] === undefined) {
                viewState[this.viewProperties[i].Company_ID] = {};
            }
            if (viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName] === undefined) {
                viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName] = {};
            }
            if (viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] === undefined) {
                if (this.viewProperties[i].DataType === "boolean") {
                    if (this.viewProperties[i].Value === "1") {
                        viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = true;
                    }
                    else {
                        viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = false;
                    }
                }
                else if (this.viewProperties[i].DataType === "number") {
                    let num = this.viewProperties[i].Value;
                    viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = num;
                }
                else {
                    viewState[this.viewProperties[i].Company_ID][this.viewProperties[i].PropertyName][this.viewProperties[i].Key] = this.viewProperties[i].Value;
                }
            }
            //#region symbol properties
            if (symbolInstanceId !== undefined && symbolInstanceId !== null && symbolInstanceId > 0) {
                for (var j = 0; j < this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())].length; j++) {
                    //note: property name
                    if (viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName] === undefined) {
                        viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName] = {};
                    }
                    //note: property key
                    if (viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] === undefined) {
                        viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = {};
                    }
                    //note: we found a property on the symbol that we should override on standard
                    if (this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].DataType === "boolean") {
                        if (this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Value === "1") {
                            viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = true;
                        }
                        else {
                            viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = false;
                        }
                    }
                    else if (this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].DataType === "number") {
                        let num = this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Value;
                        viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = num;
                    }
                    else {
                        viewState[this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Company_ID][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].PropertyName][this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Key] = this.cachedViewPropertiesSymbol[(viewName + symbolInstanceId.toString())][j].Value;
                    }
                }
            }
            //#endregion symbol properties
        }
        return viewState;
    }
    //note: initialise rules
    InitialiseCompanyRules(viewRules, viewName) {
        if ((viewRules === undefined) || (viewRules === null)) {
            return;
        }
        for (var j = 0; j < this.loginService.dataContext.length; j++) {
            if (this.loginService.dataContext[j].CompanyID === "") {
                continue;
            }
            //note: add the company
            if (this.cachedViewRules[this.loginService.dataContext[j].CompanyID] === undefined) {
                this.cachedViewRules[this.loginService.dataContext[j].CompanyID] = {};
            }
            //note: add the view
            if (this.cachedViewRules[this.loginService.dataContext[j].CompanyID][viewName] === undefined) {
                this.cachedViewRules[this.loginService.dataContext[j].CompanyID][viewName] = {};
            }
        }
        //loop and add to the cached company rules
        for (var i = 0; i < viewRules.length; i++) {
            //note: add the rule
            if (this.cachedViewRules[viewRules[i].Company_ID][viewName][viewRules[i].Name] === undefined) {
                let rule = new Rule();
                rule.Name = viewRules[i].Name;
                rule.Sequence = viewRules[i].Sequence;
                rule.ShouldAbortError = viewRules[i].ShouldAbortError;
                rule.ShouldAbortFailure = viewRules[i].ShouldAbortFailure;
                rule.ShouldRollbackOnError = viewRules[i].ShouldRollbackOnError;
                rule.ShouldRollbackOnFailure = viewRules[i].ShouldRollbackOnFailure;
                if (viewRules[i].SystemHook !== null) {
                    rule.SystemHook = this.GetSystemHooksFromStringArray(viewRules[i].SystemHook.split(','));
                }
                else {
                    rule.SystemHook = null;
                }
                rule.TriggerProperties = viewRules[i].TriggerProperties;
                rule.RuleScript = viewRules[i].RuleScript;
                rule.TrueResultScript = viewRules[i].TrueResultScript;
                rule.FalseResultScript = viewRules[i].FalseResultScript;
                rule.ViewStateRule = viewRules[i].ViewStateRule;
                rule.ModelStateRule = viewRules[i].ModelStateRule;
                rule.FriendlyName = viewRules[i].FriendlyName;
                rule.Active = viewRules[i].Active;
                rule.Description = viewRules[i].Description;
                if (viewRules[i].ExecuteRules !== null) {
                    rule.ExecuteRules = viewRules[i].ExecuteRules.split(',');
                }
                else {
                    rule.ExecuteRules = null;
                }
                this.cachedViewRules[viewRules[i].Company_ID][viewName][viewRules[i].Name] = rule;
            }
        }
    }
    //note: initialise notification
    InitialiseCompanyNotifications(viewNotifications, viewName) {
        //loop and add to the cached company Notifications
        for (var i = 0; i < viewNotifications.length; i++) {
            //note: add the company
            if (this.cachedViewNotifications[viewNotifications[i].Company_ID] === undefined) {
                this.cachedViewNotifications[viewNotifications[i].Company_ID] = {};
            }
            //note: add the view
            if (this.cachedViewNotifications[viewNotifications[i].Company_ID][viewName] === undefined) {
                this.cachedViewNotifications[viewNotifications[i].Company_ID][viewName] = {};
            }
            //note: add the notification message
            if (this.cachedViewNotifications[viewNotifications[i].Company_ID][viewName][viewNotifications[i].NotificationId] === undefined) {
                let notificationMessage = new NotificationMessage();
                notificationMessage.ErrorMessage = viewNotifications[i].ErrorMessage;
                notificationMessage.Id = viewNotifications[i].NotificationId;
                notificationMessage.IsError = viewNotifications[i].IsError;
                notificationMessage.IsWarning = viewNotifications[i].IsWarning;
                notificationMessage.IsInformation = viewNotifications[i].IsInformation;
                this.cachedViewNotifications[viewNotifications[i].Company_ID][viewName][viewNotifications[i].NotificationId] = notificationMessage;
            }
        }
    }
    //note: get system hook enum array
    GetSystemHooksFromStringArray(systemHooksString) {
        let systemHooks = [];
        for (var i = 0; i < systemHooksString.length; i++) {
            systemHooks.push(SystemHooks[systemHooksString[i]]);
        }
        return systemHooks;
    }
    //note: route to the correct view 
    Navigate(view, target) {
        if (this.cachedViewProperties[view.InternalName] === undefined) {
            this.GetPropertyConfiguration(view.InternalName).subscribe((result) => {
                this.cachedViewProperties[view.InternalName] = result;
                this.viewProperties = result;
                //note: load the rules for this view
                if (this.cachedViewRules[view.InternalName] === undefined) {
                    this.GetRules(view.InternalName).subscribe((result) => {
                        this.InitialiseCompanyRules(result, view.InternalName);
                        //note: load the notifications for this view
                        if (this.cachedViewNotifications[view.InternalName] === undefined) {
                            this.GetNotificationMessages(view.InternalName).subscribe((result) => {
                                this.InitialiseCompanyNotifications(result, view.InternalName);
                                if (target === '') {
                                    this.router.navigate([view.Route], { state: { viewFriendlyName: view.Name, internalName: view.InternalName } });
                                }
                                if (target === 'leftregion') {
                                    this.router.navigate([{ outlets: { leftregion: [view.Route] } }], { state: { viewFriendlyName: view.Name, internalName: view.InternalName } });
                                }
                            });
                        }
                    });
                }
            });
        }
        else {
            if (target === '') {
                this.router.navigate([view.Route], { state: { viewFriendlyName: view.Name, internalName: view.InternalName } });
            }
            if (target === 'leftregion') {
                this.router.navigate([{ outlets: { leftregion: [view.Route] } }], { state: { viewFriendlyName: view.Name, internalName: view.InternalName } });
            }
        }
    }
}
InitialisationService.decorators = [
    { type: Injectable }
];
InitialisationService.ctorParameters = () => [
    { type: LoginService },
    { type: HttpClient },
    { type: AppGlobal },
    { type: Router }
];

class NotificationService extends BaseService {
    constructor(initialisationService) {
        super();
        this.initialisationService = initialisationService;
        this.notificationMessages = {};
    }
    //note: given a notification id, return it's info
    GetNotification(notificationId, notificationVariables, companyID, name) {
        let notificationMessage = new NotificationMessage();
        //Object.assign(notificationMessage, this.notificationMessages[companyID][notificationId])
        Object.assign(notificationMessage, this.initialisationService.cachedViewNotifications[companyID][name][notificationId]);
        let keys = Object.keys(notificationVariables);
        notificationMessage.ErrorMessage = this.ReplaceVariables(notificationMessage.ErrorMessage, notificationVariables, keys);
        return notificationMessage;
    }
    //note: replace variables in message
    ReplaceVariables(errorMessage, notificationVariables, keys) {
        for (var i = 0; i < keys.length; i++) {
            errorMessage = errorMessage.replace(("~~" + keys[i]), notificationVariables[keys[i]]);
        }
        return errorMessage;
    }
    //note: return all notifications linked to a view for a Company
    //this may be a combination of messages from the database, but also maybe locally saved? To decrease data req...
    GetNotifications(viewName) {
        // this.notificationMessages = {
        //   QT: {
        //     1: {
        //       Id: 1,
        //       ErrorMessage: "You are exceeding the maximum allowed characters by ~~NumberCharacters",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     },      
        //     2: {
        //       Id: 2,
        //       ErrorMessage: "You require ~~NumberCharacters more characters",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     },  
        //     3: {
        //       Id: 3,
        //       ErrorMessage: "This is a required field",
        //       IsError: false,
        //       IsWarning: false,
        //       IsInformation: true        
        //     },     
        //     4: {
        //       Id: 4,
        //       ErrorMessage: "The minimum value allowed is ~~MinimumValue",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     }, 
        //     5: {
        //       Id: 5,
        //       ErrorMessage: "The maximum value allowed is ~~MaximumValue",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     },                            
        //     100000: {
        //       Id: 100000,
        //       ErrorMessage: "Description may only be 30 characters",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     },
        //   },
        //   QMC: {
        //     1: {
        //       Id: 1,
        //       ErrorMessage: "You are exceeding the maximum allowed characters by ~~NumberCharacters",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     },      
        //     2: {
        //       Id: 2,
        //       ErrorMessage: "You require ~~NumberCharacters more characters",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     },  
        //     3: {
        //       Id: 3,
        //       ErrorMessage: "This is a required field",
        //       IsError: false,
        //       IsWarning: false,
        //       IsInformation: true        
        //     },     
        //     4: {
        //       Id: 4,
        //       ErrorMessage: "The minimum value allowed is ~~MinimumValue",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     }, 
        //     5: {
        //       Id: 5,
        //       ErrorMessage: "The maximum value allowed is ~~MaximumValue",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     },                            
        //     100000: {
        //       Id: 100000,
        //       ErrorMessage: "Description may only be 30 characters",
        //       IsError: true,
        //       IsWarning: false,
        //       IsInformation: false        
        //     },
        //   }      
        // };
    }
}
NotificationService.decorators = [
    { type: Injectable }
];
NotificationService.ctorParameters = () => [
    { type: InitialisationService }
];

class Interceptor {
    constructor(loginService, appManagementService, dataContextService, notificationService) {
        this.loginService = loginService;
        this.appManagementService = appManagementService;
        this.dataContextService = dataContextService;
        this.notificationService = notificationService;
        this.refreshSubject = new Subject();
    }
    //note: fired on every http request
    intercept(req, next) {
        //note: here we have to check what is the base uri that is targeted
        //      obtain the correct token to append (since each data context will have it's own token)
        let authToken = null;
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            //note: check if we find a datacontext for the company specified in the header
            if (req.headers.get("CompanyID") === this.loginService.dataContext[i].CompanyID) {
                authToken = this.loginService.dataContext[i].Token;
                //note: token for datacontext is null for some reason
                //#region no token
                if (authToken === null) {
                    return next.handle(req)
                        .pipe(tap((event) => {
                        if (event instanceof HttpResponse) {
                        }
                        return event;
                    }));
                }
                //#endregion no token 
                //note: targeting the authorisation web service
                //#region authorisation service
                if (this.loginService.dataContext[i].CompanyID === '') {
                    const authReq = req.clone({
                        headers: req.headers.set('Authorization', 'Bearer ' + authToken)
                    });
                    return next.handle(authReq)
                        .pipe(tap((event) => {
                        if (event instanceof HttpResponse) {
                        }
                        return event;
                    }));
                }
                //#endregion authorisation service
                //note: targeting the specific datacontext web service
                else {
                    let authReq;
                    authReq = req.clone({
                        headers: req.headers.set('ApiToken', authToken)
                    });
                    //note: we will use this header value to pass any additional information to the web api service
                    if (this.appManagementService.extraHeaderValue) {
                        authReq = authReq.clone({
                            headers: authReq.headers.set('ExtraHeaderValue', this.appManagementService.extraHeaderValue)
                        });
                    }
                    return next.handle(authReq)
                        .pipe(retry(1), //note: retry each failed call by default once...this will help with bad networks etc.
                    tap((event) => {
                        if (event instanceof HttpResponse) {
                            //note: successful execution of http request
                            //      clear any errors that we previously had for this company/context
                        }
                        if (event instanceof HttpErrorResponse) {
                        }
                        return event;
                    }), catchError((error, caught) => {
                        let companyID = req.headers.get("CompanyID");
                        let dataContext = this.dataContextService.GetDataContextForCompanyId(companyID);
                        //note: our interceptor should replace the error handling
                        //      in the base service. it will also check for token
                        //      epiry. If token has expired it will use the refesh token
                        //      to obtain a new authorisation token.
                        if (error instanceof HttpErrorResponse) {
                            //note: request from the header the reason code 
                            let invalidReasonCode = error.headers.get("InvalidReasonCode"); //E: Expired
                            if (error.status
                                &&
                                    error.status === 401 //note: Not Authorised
                                &&
                                    invalidReasonCode === "E") {
                                //note: we don't want to raise an error here, just ignore
                                //      and let the refresh token take care of the issue
                                // let notificationMessage: NotificationMessage = new NotificationMessage();
                                // notificationMessage.Id = 0;
                                // notificationMessage.IsError = true;
                                // notificationMessage.ErrorMessage = error.message;
                                // this.appManagementService.globalErrors.push(notificationMessage);
                                for (var i = 0; i < this.loginService.dataContext.length; i++) {
                                    if (this.loginService.dataContext[i].CompanyID === companyID) {
                                        return this.handleTokenExpired(this.loginService.dataContext[i], i).pipe(switchMap(() => {
                                            return next.handle(this.updateHeader(req, i));
                                        }));
                                    }
                                }
                            }
                            else if ((error.status !== undefined)
                                &&
                                    error.status === 0 //note: net::ERR_CONNECTION_REFUSED...typically 1. web service offline 2. no network                 
                            ) {
                                //note: inform the datacontext
                                //note: log to our global error list so the user receives information
                                //      about the error
                                let notificationVariable = new NotificationVariable();
                                notificationVariable.DataContextDescription = dataContext.Description;
                                this.appManagementService.AddGlobalError(this.notificationService.GetNotification(10, notificationVariable, companyID, "system"));
                                //note: return EMPTY will not emit anything
                                //      forkjoin requires something to be emitted.
                                //      this is important where one of our forkjoin streams contain 
                                //      errors, but the rest not
                                //  
                                //      impersonate a successful response
                                return of(new HttpResponse({ status: 200 }));
                            }
                            else if ((error.status !== undefined)
                                &&
                                    error.status === 500 //note: Internal Server Error e.g. any backend/qmuzik error              
                            ) {
                                //note: this must be thrown so that the we can display relevant error in the view
                                //      for now, this is the only place I will throw the error
                                return throwError(error);
                            }
                            else {
                                let notificationMessage = new NotificationMessage();
                                notificationMessage.Id = 0;
                                notificationMessage.IsError = true;
                                notificationMessage.ErrorMessage = error.message;
                                this.appManagementService.globalErrors.push(notificationMessage);
                                //note: we have an HttpErrorResponse, but it is not expiry
                                //      what to do?
                                return of(new HttpResponse({ status: 200 }));
                            }
                        }
                        else {
                            //note: this is not an http error
                            let notificationMessage = new NotificationMessage();
                            notificationMessage.Id = 0;
                            notificationMessage.IsError = true;
                            notificationMessage.ErrorMessage = error.message;
                            this.appManagementService.globalErrors.push(notificationMessage);
                            return of(new HttpResponse({ status: 200 }));
                        }
                        //note: log the error as a global error
                        //      we will handle all http errors ourselves 
                        //      return success, e.g. handled
                        let notificationMessage = new NotificationMessage();
                        notificationMessage.Id = 0;
                        notificationMessage.IsError = true;
                        notificationMessage.ErrorMessage = error.message;
                        this.appManagementService.globalErrors.push(notificationMessage);
                        return of(new HttpResponse({ status: 200 }));
                    }));
                }
            }
        }
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (req.url.indexOf(this.loginService.dataContext[i].ServiceUri) > -1) {
                //found:
                authToken = this.loginService.dataContext[i].Token;
                if (authToken === null) {
                    return next.handle(req)
                        .pipe(tap((event) => {
                        if (event instanceof HttpResponse) {
                        }
                        return event;
                    }));
                }
                if (this.loginService.dataContext[i].CompanyID === '') {
                    const authReq = req.clone({
                        headers: req.headers.set('Authorization', 'Bearer ' + authToken)
                    });
                    return next.handle(authReq)
                        .pipe(tap((event) => {
                        if (event instanceof HttpResponse) {
                        }
                        return event;
                    }));
                }
                else {
                    const authReq = req.clone({
                        headers: req.headers.set('ApiToken', authToken)
                    });
                    return next.handle(authReq)
                        .pipe(tap((event) => {
                        if (event instanceof HttpResponse) {
                        }
                        return event;
                    }));
                }
            }
        }
        //nothing found: re-route to the login page (maybe with message)
        //alert('No data context found for: ' + req.url);
        return next.handle(req);
    }
    //note: this is called when a token has expired and needs to be refreshed
    handleTokenExpired(dataContext, i) {
        this.refreshSubject.subscribe({
            complete: () => {
                this.refreshSubject = new Subject();
            }
        });
        if (this.refreshSubject.observers.length === 1) {
            //note: refresh the token
            this.loginService.getTokenWithRefreshToken(dataContext).subscribe({
                next: ((i, result) => {
                    //note: here we must update the datacontext so that future api calls will use the new token
                    if (result !== null) {
                        this.loginService.dataContext[i].Token = result.Token;
                        this.loginService.dataContext[i].RefreshToken = result.RefreshToken;
                        //note: successfully retrieved the token. 
                        //      this re-issues the httprequest (switchmap)
                        this.refreshSubject.next();
                    }
                    else {
                        //note: we could not get a refresh token for some reason
                        //      the auth token is not invalid and the data context must be disabled
                        let notificationVariable = new NotificationVariable();
                        notificationVariable.DataContextDescription = dataContext.Description;
                        this.appManagementService.AddGlobalError(this.notificationService.GetNotification(11, notificationVariable, this.loginService.dataContext[i].CompanyID, "system"));
                        this.loginService.DisableDatacontext(this.loginService.dataContext[i], this.notificationService.GetNotification(12, {}, dataContext.CompanyID, "system"));
                    }
                }).bind(undefined, i),
                error: (result) => {
                    //note: an error occurred trying to get the new token with the refresh token
                    let notificationVariable = new NotificationVariable();
                    notificationVariable.DataContextDescription = dataContext.Description;
                    this.appManagementService.AddGlobalError(this.notificationService.GetNotification(11, notificationVariable, this.loginService.dataContext[i].CompanyID, "system"));
                    this.loginService.DisableDatacontext(this.loginService.dataContext[i], this.notificationService.GetNotification(12, {}, dataContext.CompanyID, "system"));
                },
                complete: () => {
                    //note: this resets the refreshsubject
                    this.refreshSubject.complete();
                }
            });
        }
        return this.refreshSubject;
    }
    //note: update the header with the new token before being sent
    updateHeader(req, i) {
        const authToken = this.loginService.dataContext[i].Token;
        //#region don't delete
        //const clonedRequest = req.clone({ setHeaders: { Authorization: 'Bearer 123' } });
        //#endregion don't delete
        req = req.clone({
            headers: req.headers.set('ApiToken', authToken)
        });
        return req;
    }
}
Interceptor.decorators = [
    { type: Injectable }
];
Interceptor.ctorParameters = () => [
    { type: LoginService },
    { type: AppManagementService },
    { type: DataContextService },
    { type: NotificationService }
];

class UsersService extends BaseService {
    constructor(http, globals) {
        super();
        this.http = http;
        this.globals = globals;
    }
    //get the logged in user by using his email address
    GetUserByEmailAddress(email) {
        return this.http.get(this.globals.apiURL + "api/Users/GetUserByEmailAddress", {
            params: {
                email: email
            }
        })
            .pipe(catchError(this.handleError));
    }
}
UsersService.decorators = [
    { type: Injectable }
];
UsersService.ctorParameters = () => [
    { type: HttpClient },
    { type: AppGlobal }
];

//this service is a utility service
//add any generic logic that can help speed up development
class UtilityService extends BaseService {
    constructor() {
        super();
        this.reloadRulesSubscription = new Subject();
        this.reloadPropertySettingsSubscription = new Subject();
        this.headerChangedSubscription = new Subject();
        //subscription: maintain a list of all open component bases
        //they have to subscribe to each other's publish events in order to communicate
        //public openComponents: ComponentBase[] = []
        this.openComponents = [];
    }
    //paging: return the requested page in an array of objects
    getPage(originalArray, pageQuantity, pageNumber) {
        let pagedArray;
        pageNumber = pageNumber - 1;
        pagedArray = originalArray.slice((pageNumber * pageQuantity), (pageNumber * pageQuantity) + pageQuantity);
        return pagedArray;
    }
}
UtilityService.decorators = [
    { type: Injectable }
];
UtilityService.ctorParameters = () => [];

class AuthGuardService {
    constructor(loginService, router) {
        this.loginService = loginService;
        this.router = router;
    }
    canActivate(route, state) {
        //if not logged in, must be routed to the login page
        if (this.loginService.isLoggedIn) {
            return true;
        }
        else {
            this.router.navigate(['/']);
        }
        return false;
    }
}
AuthGuardService.decorators = [
    { type: Injectable }
];
AuthGuardService.ctorParameters = () => [
    { type: LoginService },
    { type: Router }
];

//note: any instance related services to be hosted here
//      generic things required for processes
class InstanceService extends BaseService {
    constructor(http, loginService, router, initialisationService) {
        super();
        this.http = http;
        this.loginService = loginService;
        this.router = router;
        this.initialisationService = initialisationService;
        this.activitiesSubscription = new Subject();
        this.variableInputOutputSubscription = new Subject();
        this.symbolInstanceSubscription = new Subject();
        this.instanceVariableSubscription = new Subject();
        this.updateTaskListSubject = new Subject();
    }
    ngOnDestroy() {
    }
    /**
     * Return all activities for a symbol instance
     * @param symbolInstanceId
     */
    GetActivitiesForInstance(symbolInstanceId) {
        let observables = [];
        if (this.loginService.selectedDataContext.CompanyID === "") {
            return;
        }
        let headers = new HttpHeaders({
            "CompanyID": this.loginService.selectedDataContext.CompanyID,
            "InstanceInformation": ['0', symbolInstanceId.toString(), '0', '', 'false', '0']
        });
        let observable = this.http.get(this.loginService.selectedDataContext.ServiceUri + "api/InstanceUtility/GetActivitiesForInstance", {
            headers: headers
        })
            .pipe(catchError(this.handleError));
        observable.subscribe((result) => {
            this.activitiesSubscription.next({ "activities": result, "symbolInstanceId": symbolInstanceId });
        });
    }
    /**
     * Create a new instance and open if found
     * @param symbolInstanceId
     */
    GetNewInstance(templateDefaultAssociationId) {
        let observables = [];
        if (this.loginService.selectedDataContext.CompanyID === "") {
            return;
        }
        let headers = new HttpHeaders({
            "CompanyID": this.loginService.selectedDataContext.CompanyID,
            "InstanceInformation": ['0', '0', '0', '', 'false', '0']
        });
        let httpParams = new HttpParams();
        httpParams = httpParams.append('templateDefaultAssociationId', templateDefaultAssociationId);
        let observable = this.http.get(this.loginService.selectedDataContext.ServiceUri + "api/InstanceUtility/GetNewInstance", {
            headers: headers,
            params: httpParams
        })
            .pipe(catchError(this.handleError));
        observable.subscribe((result) => {
            this.Navigate(result.ViewName, result.InstanceId, result.SymbolInstanceId, result.InstanceMode);
        });
    }
    /**
     * Check if a qt instance should be opened on login/context switch
     */
    GetOpenOnStartup() {
        let observables = [];
        if (this.loginService.selectedDataContext.CompanyID === "") {
            return;
        }
        let headers = new HttpHeaders({
            "CompanyID": this.loginService.selectedDataContext.CompanyID,
            "InstanceInformation": ['0', '0', '0', '', 'false', '0']
        });
        let observable = this.http.get(this.loginService.selectedDataContext.ServiceUri + "api/InstanceUtility/GetOpenOnStartup", {
            headers: headers
        })
            .pipe(catchError(this.handleError));
        observable.subscribe((result) => {
            if (result !== null) {
                this.Navigate(result.ViewName, result.InstanceId, result.SymbolInstanceId, result.InstanceMode);
            }
        });
    }
    /**
     * Retrieve the symbol instance object we are working with
     * @param symbolInstanceId
     */
    GetSymbolInstance(symbolInstanceId) {
        let observables = [];
        if (this.loginService.selectedDataContext.CompanyID === "") {
            return;
        }
        let headers = new HttpHeaders({
            "CompanyID": this.loginService.selectedDataContext.CompanyID,
            "InstanceInformation": ['0', symbolInstanceId.toString(), '0', '', 'false', '0']
        });
        let observable = this.http.get(this.loginService.selectedDataContext.ServiceUri + "api/InstanceUtility/GetSymbolInstance", {
            headers: headers
        })
            .pipe(catchError(this.handleError));
        observable.subscribe((result) => {
            this.symbolInstanceSubscription.next(result);
        });
    }
    /**
     * Return all input output definitions for a symbol instance
     *
     * @param symbolInstanceId
     */
    GetVariableInputOutputForSymbolInstance(symbolInstanceId) {
        let observables = [];
        if (this.loginService.selectedDataContext.CompanyID === "") {
            return;
        }
        let headers = new HttpHeaders({
            "CompanyID": this.loginService.selectedDataContext.CompanyID,
            "InstanceInformation": ['0', symbolInstanceId.toString(), '0', '', 'false', '0']
        });
        let observable = this.http.get(this.loginService.selectedDataContext.ServiceUri + "api/InstanceUtility/GetVariableInputOutputForSymbolInstance", {
            headers: headers
        })
            .pipe(catchError(this.handleError));
        observable.subscribe((result) => {
            this.variableInputOutputSubscription.next({ "variableInputOutput": result, "symbolInstanceId": symbolInstanceId });
        });
    }
    /**
     * Return all the instance variables
     *
     * @param instanceId
     */
    GetVariablesForInstance(instanceId) {
        let observables = [];
        if (this.loginService.selectedDataContext.CompanyID === "") {
            return;
        }
        let headers = new HttpHeaders({
            "CompanyID": this.loginService.selectedDataContext.CompanyID,
            "InstanceInformation": [instanceId.toString(), '0', '0', '', 'false', '0']
        });
        let observable = this.http.get(this.loginService.selectedDataContext.ServiceUri + "api/InstanceUtility/GetVariablesForInstance", {
            headers: headers
        })
            .pipe(catchError(this.handleError));
        observable.subscribe((result) => {
            this.instanceVariableSubscription.next(result);
        });
    }
    /**
     * Route to the correct view based upon the view name
     *
     * @param viewName
     * @param instanceId
     * @param symbolInstanceId
     * @param instanceMode
     */
    Navigate(viewName, instanceId, symbolInstanceId, instanceMode) {
        //note: first identify the route based upon the view internal name 
        let view = null;
        for (var i = 0; i < Object.keys(this.initialisationService.companyViews[this.loginService.selectedDataContext.CompanyID]).length; i++) {
            let checkView = (this.initialisationService.companyViews[this.loginService.selectedDataContext.CompanyID][Object.keys(this.initialisationService.companyViews[this.loginService.selectedDataContext.CompanyID])[i]]);
            //note we have found the match
            if (checkView.InternalName === viewName) {
                view = checkView;
                break;
            }
        }
        if (view !== null) {
            this.router.navigate(['/blank']).then(() => this.router.navigate([view.Route], {
                state: {
                    viewFriendlyName: view.Name,
                    internalName: view.InternalName,
                    InstanceId: instanceId,
                    SymbolInstanceId: symbolInstanceId,
                    InstanceMode: instanceMode
                }
            }));
        }
    }
    /**
     * Handle all instance actions when a component does not inherit from componentbase
     *
     * @param modelState
     * @param isNew
     * @param instanceInformation
     */
    DoInstanceActions(modelState, isNew, instanceInformation) {
        return new Observable(null);
    }
    /**
      * Delete/Cancel an instance
      *
      * Effectively, this means setting Active=0
      *
      * Template Type will dicate if all instance history is destroyed.
      *
      * @param instanceId
      * @param symbolInstanceId
      */
    DeleteInstance(instanceId, symbolInstanceId) {
        let observables = [];
        if (this.loginService.selectedDataContext.CompanyID === "") {
            return;
        }
        let headers = new HttpHeaders({
            "CompanyID": this.loginService.selectedDataContext.CompanyID,
            "InstanceInformation": [instanceId.toString(), '0', '0', '', 'false', '0']
        });
        let httpParams = new HttpParams();
        httpParams = httpParams.append('instanceId', instanceId.toString());
        let observable = this.http.delete(this.loginService.selectedDataContext.ServiceUri + "api/InstanceUtility/DeleteInstance", {
            headers: headers,
            params: httpParams
        })
            .pipe(catchError(this.handleError));
        observable.subscribe(() => {
            this.updateTaskListSubject.next({
                "Action": "Remove",
                "InputTask": null,
                "SymbolInstanceId": symbolInstanceId
            });
        });
    }
}
InstanceService.decorators = [
    { type: Injectable }
];
InstanceService.ctorParameters = () => [
    { type: HttpClient },
    { type: LoginService },
    { type: Router },
    { type: InitialisationService }
];

class DynamicViewService extends BaseService {
    constructor(initialisationService, componentFactoryResolver, injector, compiler) {
        super();
        this.initialisationService = initialisationService;
        this.componentFactoryResolver = componentFactoryResolver;
        this.injector = injector;
        this.compiler = compiler;
        this.dynamicComponentRegistry = [];
    }
    GetDynamicComponent(internalName) {
        return __awaiter(this, void 0, void 0, function* () {
            const registryItem = this.getRegistryItemForView(internalName);
            if (registryItem == null) {
                return this.getDynamicContentErrorComponent(`Unable to find a registry item for view: ${internalName}.`);
            }
            return registryItem.platform.loadChildren().then(moduleOrFactory => {
                try {
                    if (moduleOrFactory instanceof NgModuleFactory) {
                        return moduleOrFactory;
                    }
                    else {
                        return this.compiler.compileModuleAsync(moduleOrFactory);
                    }
                }
                catch (error) {
                    return this.getDynamicContentErrorComponent(`Unable to load module ${registryItem.modulePath}.
                      Looked up using component: ${registryItem.componentName}. Error Details: ${error.message}`);
                }
            }).then((moduleFactory) => {
                try {
                    const moduleReference = moduleFactory.create(this.injector);
                    const componentResolver = moduleReference.componentFactoryResolver;
                    const componentType = moduleFactory.moduleType.dynamicComponentsMap[registryItem.componentName];
                    const componentFactory = componentResolver.resolveComponentFactory(componentType);
                    return componentFactory.create(this.injector);
                }
                catch (error) {
                    return this.getDynamicContentErrorComponent(`Unable to load module ${registryItem.modulePath}.
                      Looked up using component: ${registryItem.componentName}. Error Details: ${error.message}`);
                }
            });
        });
    }
    GetModulePathFromRegistryItem(registryItem) {
        if (registryItem && registryItem.modulePath) {
            return `${registryItem.modulePath}#${registryItem.moduleName}`;
        }
        return null;
    }
    getRegistryItemForView(internalName) {
        const registryItem = this.dynamicComponentRegistry.find(i => i.internalName === internalName);
        if (registryItem != null) {
            return registryItem;
        }
        return null;
    }
    getDynamicContentErrorComponent(errorMessage) {
        const factory = this.componentFactoryResolver.resolveComponentFactory(DynamicContentErrorComponent);
        const componentRef = factory.create(this.injector);
        const instance = componentRef.instance;
        instance.errorMessage = errorMessage;
        return componentRef;
    }
    //simulates the resolver's work when using routes to access functions
    simulateResolver(view) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.initialisationService.cachedViewRules[view.InternalName] ===
                undefined) {
                yield this.initialisationService
                    .GetRules(view.InternalName)
                    .toPromise()
                    .then((result) => {
                    this.initialisationService.InitialiseCompanyRules(result, view.InternalName);
                });
            }
            if (this.initialisationService.cachedViewNotifications[view.InternalName] ===
                undefined) {
                yield this.initialisationService
                    .GetNotificationMessages(view.InternalName)
                    .toPromise()
                    .then((result) => {
                    this.initialisationService.InitialiseCompanyNotifications(result, view.InternalName);
                });
            }
            if (this.initialisationService.cachedViewProperties[view.InternalName] ===
                undefined) {
                yield this.initialisationService
                    .GetPropertyConfiguration(view.InternalName)
                    .toPromise()
                    .then((result) => {
                    this.initialisationService.cachedViewProperties[view.InternalName] = result;
                });
            }
            return true;
        });
    }
}
DynamicViewService.decorators = [
    { type: Injectable }
];
DynamicViewService.ctorParameters = () => [
    { type: InitialisationService },
    { type: ComponentFactoryResolver },
    { type: Injector },
    { type: Compiler }
];

/**
 * Services for the component options component
 */
class ComponentOptionsService extends BaseService {
    constructor(dynamicViewService, loginService, initialisationService, changeDetectorRef) {
        super();
        this.dynamicViewService = dynamicViewService;
        this.loginService = loginService;
        this.initialisationService = initialisationService;
        this.changeDetectorRef = changeDetectorRef;
        this.modelstateProperties = [];
        this.numericModelStateProperties = [];
        this.sortProperties = [];
        this.aggregateProperties = [];
        this.aggregates = [];
        this.aggregateMap = new Map();
        this.labels = new Map();
        this.modelStatePropertyLabels = new Map();
        this.modelStatePropertyLabelsLoading = false;
        this.employeeDefaults = [];
        this.changesAppliedtoAllChildren = false;
        this.saveToDefaultTrigger = new Subject();
        this.deleteDefaultTrigger = new Subject();
        this.sortListTrigger = new Subject();
        this.calcAggregateTrigger = new Subject();
        this.exportToExcelTrigger = new Subject();
        this.undoLastChangeTrigger = new Subject();
        this.undoAllChangesTrigger = new Subject();
        this.applyChangesToAllChildren = new Subject();
        this.serialisationLevelChanged = new Subject();
        this.labelsLoaded = new Subject();
        this.closeOptionsTrigger = new Subject();
        this.pinOptionsTrigger = new Subject();
        this.serializationLevel = 0;
        this.views = [];
        this.modelStatePropertyLabelsLoaded = false;
        this.setDefaultLabels();
        this.loadViews();
    }
    /**
     * Cleanup
     */
    ngOnDestroy() {
        if (this.initialisationServiceSubscription) {
            this.initialisationServiceSubscription.unsubscribe();
        }
    }
    setSerialisationLevel(level) {
        this.serializationLevel = level;
        this.serialisationLevelChanged.next(this.serializationLevel);
    }
    /**
     * Extract stuff from component base
     *
     * @param properties
     */
    loadModelStateProperties(properties) {
        this.modelstateProperties.push(...properties);
    }
    loadNumericModelStateProperties(numericProperties) {
        this.numericModelStateProperties.push(...numericProperties);
    }
    /**
     * Sort
     */
    sortModelStateList() {
        this.sortListTrigger.next(this.sortProperties);
    }
    /**
     * Aggregate mechanism
     *
     * @param propertyAction
     */
    calculateAggregate(propertyAction) {
        this.calcAggregateTrigger.next(propertyAction);
    }
    /**
     * Save to defaults
     *
     * @param defaultName
     */
    saveToDefaults(defaultName) {
        this.saveToDefaultTrigger.next(defaultName);
    }
    deleteDefault(defaultInstance) {
        this.deleteDefaultTrigger.next(defaultInstance);
    }
    setApplyToAll(value) {
        this.changesAppliedtoAllChildren = value;
        this.applyChangesToAllChildren.next(value);
    }
    getComponentOptionLabel(key) {
        return this.labels.get(key);
    }
    setLabels(labels) {
        this.labels = labels;
        this.populateAggregates();
        this.labelsLoaded.next();
    }
    exportToExcel() {
        this.exportToExcelTrigger.next();
    }
    undoLastChange() {
        this.undoLastChangeTrigger.next();
    }
    undoAllChanges() {
        this.undoAllChangesTrigger.next();
    }
    setDefaultLabels() {
        return __awaiter(this, void 0, void 0, function* () {
            this.labels.set("AggregateCount", "Count");
            this.labels.set("AggregateMin", "Minimum");
            this.labels.set("AggregateMax", "Maximum");
            this.labels.set("AggregateAvg", "Average");
            this.labels.set("AggregateSum", "Total");
            this.labels.set("ApplyToAll", "Apply changes to all");
            this.labels.set("ListOptions", "List Options");
            this.labels.set("Options", "Options");
            this.labels.set("SortList", "Sort List");
            this.labels.set("SortBy", "Sort by:");
            this.labels.set("CalcAggregate", "Calculate Aggregate(s)");
            this.labels.set("CalcAggregateFor", "Calculate Aggregate(s) for");
            this.labels.set("ChooseAggregate", "Choose...");
            this.labels.set("AggregateContext", "Context");
            this.labels.set("AggregateContextDisplayOnly", "Displayed");
            this.labels.set("AggregateContextAll", "All");
            this.labels.set("AggregateContextSelectedList", "Selected");
            this.labels.set("AggregateTHProperty", "Property");
            this.labels.set("AggregateTHAggregate", "Aggregate");
            this.labels.set("AggregateTHResult", "Result");
            this.labels.set("ManageDefaults", "Manage my defaults");
            this.labels.set("DefaultName", "Default Name:");
            this.labels.set("ExistingDefaults", "Existing default(s)");
            this.labels.set("SerializationLevel", "Serialization Level");
            this.labels.set("ExportToExcel", "Export To Excel");
            this.labels.set("Undo", "Undo");
            this.populateAggregates();
            this.labelsLoaded.next();
        });
    }
    populateAggregates() {
        return __awaiter(this, void 0, void 0, function* () {
            this.aggregates = [];
            this.aggregates.push(this.labels.get("AggregateCount"));
            this.aggregates.push(this.labels.get("AggregateMin"));
            this.aggregates.push(this.labels.get("AggregateMax"));
            this.aggregates.push(this.labels.get("AggregateAvg"));
            this.aggregates.push(this.labels.get("AggregateSum"));
            this.aggregateMap.set(this.labels.get("AggregateCount"), "COUNT");
            this.aggregateMap.set(this.labels.get("AggregateMin"), "MIN");
            this.aggregateMap.set(this.labels.get("AggregateMax"), "MAX");
            this.aggregateMap.set(this.labels.get("AggregateAvg"), "AVG");
            this.aggregateMap.set(this.labels.get("AggregateSum"), "SUM");
        });
    }
    getComponent(viewName) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.dynamicViewService.GetDynamicComponent(viewName);
        });
    }
    ViewExists(name) {
        for (var i = 0; i < this.views.length; i++) {
            if (this.views[i].Name === name) {
                //found
                return true;
            }
        }
        //not found
        return false;
    }
    GetView(internalName) {
        for (var i = 0; i < this.views.length; i++) {
            if (this.views[i].InternalName === internalName) {
                //found
                return this.views[i];
            }
        }
        //not found
        return null;
    }
    loadViews() {
        let view;
        let views = new Views();
        //note: 1st check if view have been retrieved
        //this must change, once we have decided how to handle views for different companies
        if (Object.keys(this.initialisationService.companyViews).length > 0) {
            //note: run through views returned from all companies
            for (var i = 0; i < Object.keys(this.initialisationService.companyViews).length; i++) {
                let views = this.initialisationService.companyViews[Object.keys(this.initialisationService.companyViews)[i]];
                for (var j = 0; j < Object.keys(views).length; j++) {
                    let view = views[Object.keys(views)[j]];
                    if (!this.ViewExists(view.Name)) {
                        this.views.push(view);
                    }
                }
            }
        }
        else {
            //note: set up the views for each company
            for (var i = 0; i < this.loginService.dataContext.length; i++) {
                if (this.loginService.dataContext[i].CompanyID === "") {
                    continue;
                }
                if (this.initialisationService.companyViews[this.loginService.dataContext[i].CompanyID] === undefined) {
                    this.initialisationService.companyViews[this.loginService.dataContext[i].CompanyID] = new Views();
                }
            }
            this.initialisationServiceSubscription = this.initialisationService
                .GetMyViews()
                .subscribe((result) => {
                //loop and create the view for each company
                for (var i = 0; i < result.length; i++) {
                    view = new View();
                    view.InternalName = result[i].InternalName;
                    view.Name = result[i].Name;
                    view.EntityName = result[i].EntityName;
                    view.Route = result[i].Route;
                    view.IsMenuItem = result[i].IsMenuItem;
                    this.initialisationService.companyViews[result[i].Company_ID][view.InternalName] = view;
                }
                //note: run through views returned from all companies
                for (var i = 0; i < Object.keys(this.initialisationService.companyViews).length; i++) {
                    let views = this.initialisationService.companyViews[Object.keys(this.initialisationService.companyViews)[i]];
                    for (var j = 0; j < Object.keys(views).length; j++) {
                        let view = views[Object.keys(views)[j]];
                        if (!this.ViewExists(view.Name) && view.IsMenuItem) {
                            this.views.push(view);
                        }
                    }
                }
                //note: when jumping from one component to next, we will use the activated views
                //to decide which view to open (together with the Entity property)
                this.initialisationService.activatedViews = this.views;
                //this.viewsRetrievedSubscription.next(null);
            });
        }
    }
    formatString(input, ...args) {
        args.forEach((arg, i) => {
            input = input.replace(new RegExp('\\{' + i + '\\}'), arg);
        });
        return input;
    }
    simulateResolver(view) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.initialisationService.cachedViewRules[view.InternalName] ===
                undefined) {
                yield this.initialisationService
                    .GetRules(view.InternalName)
                    .toPromise()
                    .then((result) => {
                    this.initialisationService.InitialiseCompanyRules(result, view.InternalName);
                });
            }
            if (this.initialisationService.cachedViewNotifications[view.InternalName] ===
                undefined) {
                yield this.initialisationService
                    .GetNotificationMessages(view.InternalName)
                    .toPromise()
                    .then((result) => {
                    this.initialisationService.InitialiseCompanyNotifications(result, view.InternalName);
                });
            }
            if (this.initialisationService.cachedViewProperties[view.InternalName] ===
                undefined) {
                yield this.initialisationService
                    .GetPropertyConfiguration(view.InternalName)
                    .toPromise()
                    .then((result) => {
                    this.initialisationService.cachedViewProperties[view.InternalName] = result;
                });
            }
            return true;
        });
    }
    CloseOptions() {
        this.closeOptionsTrigger.next();
    }
    ToggleOptionsPin() {
        this.pinOptionsTrigger.next();
    }
}
ComponentOptionsService.decorators = [
    { type: Injectable }
];
ComponentOptionsService.ctorParameters = () => [
    { type: DynamicViewService },
    { type: LoginService },
    { type: InitialisationService },
    { type: ChangeDetectorRef }
];

class ColumnValuesService extends BaseService {
    //private languageCodeService: LanguageCodeService;
    constructor(http, loginService, dataContextService) {
        super();
        this.http = http;
        this.loginService = loginService;
        this.dataContextService = dataContextService;
        this.defaultColumnValues = [];
        this.servicesInjected = false;
        //#endregion
        //#region header control search 
        this.formatter = (result) => (this.GetFormatter(result));
        this.inputFormatter = (result) => (result.ColumnName + " " + result.ColumnValue);
        //retrieve defaults used for creation of new model instance
        this.GetDefaults();
    }
    //#region injection
    InjectServices() {
        //this.languageCodeService = AppInjector.get(LanguageCodeService);
        this.servicesInjected = true;
    }
    //#endregion injection
    //this is how we will retrieve column values
    GetColumnValuesByColumnName(columnName) {
        let uri;
        //here we know the CompanyID
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            //column values will always be the same accross databases?? Different hotfixes??
            //use the first company found for now
            uri = this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary/GetColumn_ValuesByColumnName";
            break;
        }
        let httpParams = new HttpParams();
        httpParams = httpParams.append('columnName', columnName);
        let headers = new HttpHeaders({
            "CompanyID": this.loginService.dataContext[i].CompanyID
        });
        return this.http.get(uri, {
            params: httpParams,
            headers: headers
        })
            .pipe(catchError(this.handleError));
    }
    //#region crud
    //generated: create a new instance of the model
    CreateNew() {
        //note: default values must be obtained for each company from the rest api
        let source;
        let target = new ColumnValues();
        for (var i = 0; i < this.defaultColumnValues.length; i++) {
            if (this.defaultColumnValues[i].CompanyID == this.loginService.defaultCompanyID) {
                source = this.defaultColumnValues[i];
                break;
            }
        }
        Object.assign(target, source);
        return target;
    }
    //generated: call to rest api to delete a model instance
    Delete(modelState) {
        let uri;
        //here we know the CompanyID
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID == modelState.CompanyID) {
                uri = this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary";
                break;
            }
        }
        let httpParams = new HttpParams();
        httpParams = httpParams.append("ColumnName", modelState.ColumnName);
        httpParams = httpParams.append("ColumnValue", modelState.ColumnValue);
        let headers = new HttpHeaders({
            "CompanyID": modelState.CompanyID
        });
        return this.http.delete(uri, {
            params: httpParams,
            headers: headers
        })
            .pipe(catchError(this.handleError));
    }
    //generated: call to rest api to save a new model instance, or update and existing model instance
    Save(modelState, isNew) {
        let uri;
        modelState["PropertyStoreType"] = null;
        //here we know the CompanyID
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID == modelState.CompanyID) {
                uri = this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary";
                break;
            }
        }
        let headers = new HttpHeaders({
            "CompanyID": modelState.CompanyID
        });
        if (isNew) {
            return this.http.post(this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary/PostColumn_Values", modelState, {
                headers: headers
            })
                .pipe(catchError(this.handleError));
        }
        else {
            let httpParams = new HttpParams();
            httpParams = httpParams.append("ColumnName", modelState.ColumnName);
            httpParams = httpParams.append("ColumnValue", modelState.ColumnValue);
            return this.http.put(this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary", modelState, {
                params: httpParams,
                headers: headers
            })
                .pipe(catchError(this.handleError));
        }
    }
    //#endregion
    SaveInstance(modelState, isNew, instanceId, symbolInstanceId, symbolInstanceActivityId, chosenSymbolInstanceActivityId, instanceAction) {
        return new Observable(null);
    }
    //#region create for
    //generated: we need an instance to build our form initially
    CreateForInitialisation() {
        let columnValues = new ColumnValues();
        columnValues.CompanyID = this.loginService.defaultCompanyID;
        return columnValues;
    }
    //generated: initialise all properties of model instance to null
    CreateForFilter() {
        let columnValues = new ColumnValues();
        Object.keys(columnValues).forEach((key) => {
            columnValues[key] = null;
        });
        columnValues.CompanyID = this.loginService.defaultCompanyID;
        return columnValues;
    }
    //#endregion
    //#region assign to type
    //generated: selecting an object in the typeahead we need to "cast" it to the required type so that instanceof works
    AssignToType(item) {
        //this is just for testing
        //once we generated classes for company tables 
        //we won't have to select it here
        //item.CompanyID = this.loginService.defaultCompanyID;
        //must use assign here, otherwise instancof checks will not work
        return Object.assign(new ColumnValues(), item);
    }
    //generated: assign all foreign key properties to it's actual type so that instanceof works
    AssignToTypeForeignKey(item, foreignKeyProperty) {
        if (!this.servicesInjected) {
            this.InjectServices();
        }
        // if(foreignKeyProperty == 'RefLanguageCode'){
        //   //must use assign here, otherwise instancof checks will not work
        //   return this.languageCodeService.AssignToType(item);
        // }
        if (foreignKeyProperty == 'Ref_CompanyID') {
            //must use assign here, otherwise instancof checks will not work
            return this.dataContextService.AssignToType(item);
        }
    }
    //note: if multi-company we must display the id so a user can "know"
    GetFormatter(result) {
        if (this.loginService.defaultCompanyID === "") {
            for (var i = 0; i < this.loginService.dataContext.length; i++) {
                if (result.CompanyID === this.loginService.dataContext[i].CompanyID) {
                    return result.ColumnName + " | " + result.ColumnValue + "   |   " + this.loginService.dataContext[i].CompanyName;
                }
            }
        }
        else {
            return result.ColumnName + " | " + result.ColumnValue;
        }
    }
    //#endregion
    //generated: for all companies, retrieve the default values for a new model instance
    GetDefaults() {
        //run through data contexts and get the defaults for a company
        let observables = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary/GetColumn_ValuesDefault", {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observables.push(observable);
        }
        ;
        forkJoin(observables).pipe(map((data) => {
            return this.FormatObserveablesArray(data);
        })).subscribe((result) => {
            this.defaultColumnValues = result;
        });
    }
    //generated: return a single object by primary key
    GetByPrimaryKey(ColumnName, ColumnValue, CompanyID) {
        let uri;
        //here we know the CompanyID
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID == CompanyID) {
                uri = this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary";
                break;
            }
        }
        let httpParams = new HttpParams();
        httpParams = httpParams.append("ColumnName", ColumnName);
        httpParams = httpParams.append("ColumnValue", ColumnValue);
        let headers = new HttpHeaders({
            "CompanyID": CompanyID
        });
        return this.http.get(uri, {
            params: httpParams,
            headers: headers
        })
            .pipe(catchError(this.handleError));
    }
    Search(filter, advancedFilter) {
        //get the operator array
        //always use "like" for the standard filter
        let operatorArray;
        //so, remember why we do it this way:
        //we need to send the model state to the rest service
        //reason: depending on the model state it may filter itself differently
        //that is not something I can "generate", it depends on business rules
        let searchCriteria = {};
        searchCriteria.filter = filter;
        //add any advanced filters to the array
        if ((advancedFilter !== undefined) && (advancedFilter !== null)) {
            if (advancedFilter.length != 0) {
                searchCriteria.filter = [];
            }
            for (var j = 0; j < advancedFilter.length; j++) {
                advancedFilter[j]["HeaderControl"] = null;
                advancedFilter[j]["Ref_CompanyID"] = null;
                searchCriteria.filter.push(advancedFilter[j]);
            }
            operatorArray = this.GetOperatorArray(null, advancedFilter);
        }
        else {
            operatorArray = this.GetOperatorArray(filter, null);
        }
        searchCriteria.operators = operatorArray;
        let observables = [];
        if (this.loginService.defaultCompanyID === "") {
            for (var i = 0; i < this.loginService.dataContext.length; i++) {
                if (this.loginService.dataContext[i].CompanyID === "") {
                    continue;
                }
                let headers = new HttpHeaders({
                    "CompanyID": this.loginService.dataContext[i].CompanyID
                });
                let observable = this.http.post(this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary/PostColumnValues_HeaderControl", searchCriteria, {
                    headers: headers
                })
                    .pipe(catchError(this.handleError));
                observables.push(observable);
            }
        }
        else {
            //note: if a "company" is selected globally only search there
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.selectedDataContext.CompanyID
            });
            let observable = this.http.post(this.loginService.selectedDataContext.ServiceUri + "api/Column_ValuesLibrary/PostColumnValues_HeaderControl", searchCriteria, {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observables.push(observable);
        }
        return forkJoin(observables).pipe(map((data) => {
            return this.FormatObserveablesArray(data);
        }));
    }
    //return all without filtering
    GetColumn_Values() {
        //here is now where the shite his the fan
        //we must merge multiple observables into one result set
        //e.g. loop through all our logged in user's data contexts
        //create an observable for each of those contexts
        //merge the result
        let observables = [];
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === "") {
                continue;
            }
            let headers = new HttpHeaders({
                "CompanyID": this.loginService.dataContext[i].CompanyID
            });
            let observable = this.http.get(this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary", {
                headers: headers
            })
                .pipe(catchError(this.handleError));
            observables.push(observable);
        }
        return forkJoin(observables).pipe(map((data) => {
            return this.FormatObserveablesArray(data);
        }));
    }
    //all typeaheads used to find an entity value
    //in the database will generate their own method
    //the object model state will be sent to the api
    //since the data may be required to filter in some
    //custom way
    Get_ColumnValues_RefLanguageCode(filter, advancedFilter, modelState) {
        //get the operator array
        //always use "like" for the standard filter
        let operatorArray;
        //so, remember why we do it this way:
        //we need to send the model state to the rest service
        //reason: depending on the model state it may filter itself differently
        //that is not something I can "generate", it depends on business rules
        let searchCriteria = {};
        searchCriteria.filter = filter;
        //add any advanced filters to the array
        if ((advancedFilter !== undefined) && (advancedFilter !== null)) {
            if (advancedFilter.length != 0) {
                searchCriteria.filter = [];
            }
            for (var j = 0; j < advancedFilter.length; j++) {
                advancedFilter[j]["HeaderControl"] = null;
                advancedFilter[j]["Ref_CompanyID"] = null;
                searchCriteria.filter.push(advancedFilter[j]);
            }
            operatorArray = this.GetOperatorArray(null, advancedFilter);
        }
        else {
            operatorArray = this.GetOperatorArray(filter, null);
        }
        searchCriteria.operators = operatorArray;
        searchCriteria.modelState = {};
        Object.assign(searchCriteria.modelState, modelState);
        searchCriteria.operators = operatorArray;
        //note: if the modelstate has a company id we should only use that
        //only search on all if the company id is null or spaces
        let observables = [];
        searchCriteria.modelState.CompanyID = modelState.CompanyID;
        searchCriteria.modelState.HeaderControl = null;
        let headers = new HttpHeaders({
            "CompanyID": modelState.CompanyID
        });
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID === modelState.CompanyID) {
                break;
            }
        }
        let observable = this.http.post(this.loginService.dataContext[i].ServiceUri + "api/Language_CodeLibrary/PostColumnValues_RefLanguageCode", searchCriteria, {
            headers: headers
        })
            .pipe(catchError(this.handleError));
        observables.push(observable);
        return forkJoin(observables).pipe(map((data) => {
            return this.FormatObserveablesArray(data);
        }));
    }
    //generated: each foreign key must be loadable from here
    GetForeignKeyObject(foreignKey, values, companyID) {
        if (!this.servicesInjected) {
            this.InjectServices();
        }
        // if(foreignKey === "RefLanguageCode"){
        //   //in qmuzik foreign keys are represented by spaces
        //   if((values[0].replace(/ /g, "")) === "" ||  (values[0] === null))
        //   {
        // 	return new Observable(null);
        //   }
        //   return this.languageCodeService.GetByPrimaryKey(values[0], companyID);
        // }  
        if (foreignKey === "Ref_CompanyID") {
            //in qmuzik foreign keys are represented by spaces
            if ((values[0].replace(/ /g, "")) === "" || (values[0] === null)) {
                return new Observable(null);
            }
            for (var i = 0; i < this.loginService.dataContext.length; i++) {
                if (this.loginService.dataContext[i].CompanyID === values[0]) {
                    return of(this.loginService.dataContext[i]);
                }
            }
        }
    }
    //composite view: a foreign key value has been selected
    //we must now filter on this foreign key
    FilterByForeignKey(filterOnProperties, filterFromProperties, modelState) {
        return new Observable(null);
    }
    //generated: call to rest api to validate a property
    AsyncValidation(property, modelState, isNew) {
        let validationCriteria = {};
        validationCriteria.property = property;
        validationCriteria.modelState = modelState;
        validationCriteria.isNew = isNew;
        let uri;
        //here we know the CompanyID
        for (var i = 0; i < this.loginService.dataContext.length; i++) {
            if (this.loginService.dataContext[i].CompanyID == modelState.CompanyID) {
                uri = this.loginService.dataContext[i].ServiceUri + "api/Column_ValuesLibrary/PostAsyncValidate";
                break;
            }
        }
        let headers = new HttpHeaders({
            "CompanyID": modelState.CompanyID
        });
        return this.http.post(uri, validationCriteria, {
            headers: headers
        })
            .pipe(catchError(this.handleError));
    }
}
ColumnValuesService.decorators = [
    { type: Injectable }
];
ColumnValuesService.ctorParameters = () => [
    { type: HttpClient },
    { type: LoginService },
    { type: DataContextService }
];

//note: service

class QmuzikAngularServiceModule {
}
QmuzikAngularServiceModule.decorators = [
    { type: NgModule, args: [{
                imports: [],
                declarations: [],
                providers: [],
                exports: []
            },] }
];

/*
 * Public API Surface of qmuzik-angular-service
 */

/**
 * Generated bundle index. Do not edit.
 */

export { AppManagementService, AuthGuardService, BaseService, ColumnValuesService, ComponentOptionsService, DataContextService, DynamicViewService, InitialisationService, InstanceService, Interceptor, LoginService, NotificationService, QmuzikAngularServiceModule, UsersService, UtilityService };
//# sourceMappingURL=qmuzik-angular-service.js.map
